"""
Esmerald: Highly scalable, performant, easy to learn and for every application.
"""
__version__ = "0.5.4"

from starlette import status

from esmerald.conf import settings
from esmerald.conf.global_settings import EsmeraldAPISettings
from esmerald.injector import Inject

from .applications import ChildEsmerald, Esmerald
from .backgound import BackgroundTask, BackgroundTasks
from .config import (
    CORSConfig,
    CSRFConfig,
    JWTConfig,
    OpenAPIConfig,
    SessionConfig,
    StaticFilesConfig,
    TemplateConfig,
)
from .datastructures import UploadFile
from .exceptions import (
    HTTPException,
    ImproperlyConfigured,
    MethodNotAllowed,
    NotAuthenticated,
    NotFound,
    PermissionDenied,
    ServiceUnavailable,
    ValidationErrorException,
)
from .params import Body, Cookie, Header, Injects, Param, Path, Query
from .permissions import AllowAny, BasePermission, DenyAll
from .protocols import AsyncDAOProtocol, DaoProtocol, MiddlewareProtocol
from .requests import Request
from .responses import JSONResponse, ORJSONResponse, Response, TemplateResponse, UJSONResponse
from .routing.gateways import Gateway, WebSocketGateway
from .routing.handlers import delete, get, patch, post, put, route, websocket
from .routing.router import Include, Router
from .routing.views import APIView
from .schedulers.asyncz.handler import scheduler
from .template import JinjaTemplateEngine, MakoTemplateEngine
from .websockets import WebSocket, WebSocketDisconnect

__all__ = [
    "AllowAny",
    "APIView",
    "AsyncDAOProtocol",
    "BackgroundTask",
    "BackgroundTasks",
    "Body",
    "BasePermission",
    "ChildEsmerald",
    "CORSConfig",
    "CSRFConfig",
    "Cookie",
    "DaoProtocol",
    "DenyAll",
    "Esmerald",
    "EsmeraldAPISettings",
    "Gateway",
    "Header",
    "HTTPException",
    "Include",
    "Inject",
    "Injects",
    "ImproperlyConfigured",
    "JSONResponse",
    "JinjaTemplateEngine",
    "JWTConfig",
    "MakoTemplateEngine",
    "MethodNotAllowed",
    "MiddlewareProtocol",
    "NotAuthenticated",
    "NotFound",
    "ORJSONResponse",
    "OpenAPIConfig",
    "Param",
    "Path",
    "PermissionDenied",
    "Query",
    "Request",
    "Response",
    "Router",
    "ServiceUnavailable",
    "SessionConfig",
    "StaticFilesConfig",
    "TemplateConfig",
    "TemplateResponse",
    "UJSONResponse",
    "UploadFile",
    "ValidationErrorException",
    "WebSocket",
    "WebSocketDisconnect",
    "WebSocketGateway",
    "delete",
    "get",
    "patch",
    "post",
    "put",
    "route",
    "scheduler",
    "settings",
    "status",
    "websocket",
]



"""
Signature is widely used by Pydantic and comes from the inpect library.
A lot of great work was done using the Signature and Esmerald is no exception.
"""

from inspect import Parameter as InspectParameter
from inspect import Signature
from typing import TYPE_CHECKING, Any, ClassVar, Optional, Set, Union

from esmerald.exceptions import (
    ImproperlyConfigured,
    InternalServerError,
    ValidationErrorException,
)
from esmerald.requests import Request
from esmerald.transformers.constants import UNDEFINED
from esmerald.transformers.utils import get_connection_info
from esmerald.utils.helpers import is_optional_union
from esmerald.websockets import WebSocket
from pydantic import BaseModel, ValidationError

if TYPE_CHECKING:
    from pydantic.error_wrappers import ErrorDict
    from pydantic.typing import DictAny

IntValError = Union[InternalServerError, ValidationError]


class EsmeraldSignature(BaseModel):
    dependency_names: ClassVar[Set[str]]
    return_annotation: ClassVar[Any]

    class Config:
        arbitrary_types_allowed = True

    @classmethod
    def parse_values_for_connection(
        cls, connection: Union[Request, WebSocket], **kwargs: "DictAny"
    ) -> "DictAny":
        try:
            signature = cls(**kwargs)
            values = {}
            for key in cls.__fields__:
                values[key] = signature.field_value(key)
            return values
        except ValidationError as e:
            raise cls.build_exception(connection, e) from e

    @classmethod
    def build_exception(
        cls, connection: Union[Request, WebSocket], exception: ValidationError
    ) -> IntValError:
        server_errors = []
        client_errors = []

        for err in exception.errors():
            if not cls.is_server_error(err):
                client_errors.append(err)
            else:
                server_errors.append(err)

        method, url = get_connection_info(connection=connection)
        error_message = f"Validation failed for {url} with method {method}."

        if client_errors:
            return ValidationErrorException(detail=error_message, extra=client_errors)
        return InternalServerError(detail=error_message, extra=server_errors)

    @classmethod
    def is_server_error(cls, error: "ErrorDict") -> bool:
        """
        Classic approach functionality used widely to check if is a server error or not.
        """
        return error["loc"][-1] in cls.dependency_names

    def field_value(self, key: str) -> Any:
        return self.__getattribute__(key)


class Parameter(BaseModel):
    annotation: Optional[Any]
    default: Optional[Any]
    name: Optional[str]
    optional: Optional[bool]
    fn_name: Optional[str]
    param_name: Optional[str]
    parameter: Optional[InspectParameter]

    class Config:
        arbitrary_types_allowed = True

    def __init__(
        self, fn_name: str, param_name: str, parameter: InspectParameter, **kwargs: "DictAny"
    ) -> None:
        super().__init__(**kwargs)
        if parameter.annotation is Signature.empty:
            raise ImproperlyConfigured(
                f"The parameter name {param_name} from {fn_name} does not have a type annotation. "
                "If it should receive any value, use 'Any' as type."
            )
        self.annotation = parameter.annotation
        self.default = parameter.default
        self.param_name = param_name
        self.name = param_name
        self.optional = is_optional_union(self.annotation)

    @property
    def default_defined(self) -> bool:
        return self.default not in UNDEFINED



from typing import TYPE_CHECKING, Any

from typing_extensions import Protocol, runtime_checkable

if TYPE_CHECKING:
    from esmerald.types import ASGIApp, Receive, Scope, Send


@runtime_checkable
class MiddlewareProtocol(Protocol):  # pragma: no cover
    def __init__(self, app: "ASGIApp", **kwargs: Any):
        ...

    async def __call__(self, scope: "Scope", receive: "Receive", send: "Send") -> None:
        ...



from datetime import timezone
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncContextManager,
    Callable,
    Dict,
    List,
    Optional,
    Sequence,
    Union,
)

from asyncz.contrib.esmerald.scheduler import EsmeraldScheduler
from esmerald.conf import settings
from esmerald.config import CORSConfig, CSRFConfig, SessionConfig, TemplateConfig
from esmerald.config.openapi import OpenAPIConfig
from esmerald.config.static_files import StaticFilesConfig
from esmerald.datastructures import State
from esmerald.exception_handlers import (
    improperly_configured_exception_handler,
    validation_error_exception_handler,
)
from esmerald.exceptions import ImproperlyConfigured, ValidationErrorException
from esmerald.middleware.asyncexitstack import AsyncExitStackMiddleware
from esmerald.middleware.cors import CORSMiddleware
from esmerald.middleware.csrf import CSRFMiddleware
from esmerald.middleware.exceptions import (
    EsmeraldAPIExceptionMiddleware,
    ExceptionMiddleware,
)
from esmerald.middleware.sessions import SessionMiddleware
from esmerald.middleware.trustedhost import TrustedHostMiddleware
from esmerald.permissions.types import Permission
from esmerald.protocols.template import TemplateEngineProtocol
from esmerald.routing import gateways
from esmerald.routing.router import HTTPHandler, Include, Router, WebSocketHandler
from esmerald.types import (
    APIGateHandler,
    ASGIApp,
    Dependencies,
    ExceptionHandlers,
    LifeSpanHandler,
    Middleware,
    ParentType,
    Receive,
    ResponseCookies,
    ResponseHeaders,
    ResponseType,
    RouteParent,
    SchedulerType,
    Scope,
    Send,
)
from openapi_schemas_pydantic.v3_1_0 import License, SecurityRequirement, Server
from openapi_schemas_pydantic.v3_1_0.open_api import OpenAPI
from starlette.applications import Starlette
from starlette.middleware import Middleware as StarletteMiddleware  # noqa

if TYPE_CHECKING:
    from openapi_schemas_pydantic.v3_1_0 import SecurityRequirement


class Esmerald(Starlette):
    """
    Creates a Esmerald application instance.
    """

    __slots__ = (
        "title",
        "app_name",
        "description",
        "version",
        "contact",
        "terms_of_service",
        "license",
        "servers",
        "secret",
        "allowed_hosts",
        "allow_origins",
        "permissions",
        "dependencies",
        "middleware",
        "exception_handlers",
        "csrf_config",
        "openapi_config",
        "cors_config",
        "static_files_config",
        "template_config",
        "session_config",
        "response_class",
        "response_cookies",
        "response_headers",
        "scheduler_class",
        "scheduler",
        "tags",
        "root_path",
        "deprecated",
        "security",
        "include_in_schema",
        "redirect_slashes",
    )

    def __init__(
        self,
        *,
        debug: bool = settings.debug,
        title: Optional[str] = settings.title,
        version: Optional[str] = settings.version,
        summary: Optional[str] = settings.summary,
        app_name: Optional[str] = settings.app_name,
        description: Optional[str] = settings.description,
        contact: Optional[Dict[str, Union[str, Any]]] = settings.contact,
        terms_of_service: Optional[str] = settings.terms_of_service,
        license: Optional[License] = None,
        security: Optional[List[SecurityRequirement]] = None,
        servers: List[Server] = [Server(url="/")],
        secret_key: Optional[str] = settings.secret_key,
        allowed_hosts: Optional[List[str]] = settings.allowed_hosts,
        allow_origins: Optional[List[str]] = settings.allow_origins,
        permissions: Optional[List["Permission"]] = settings.permissions,
        dependencies: Optional["Dependencies"] = settings.dependencies,
        csrf_config: Optional["CSRFConfig"] = settings.csrf_config,
        openapi_config: Optional["OpenAPIConfig"] = settings.openapi_config,
        cors_config: Optional["CORSConfig"] = settings.cors_config,
        static_files_config: Optional["StaticFilesConfig"] = settings.static_files_config,
        template_config: Optional["TemplateConfig"] = settings.template_config,
        session_config: Optional["SessionConfig"] = settings.session_config,
        response_class: Optional["ResponseType"] = settings.response_class,
        response_cookies: Optional["ResponseCookies"] = settings.response_cookies,
        response_headers: Optional["ResponseHeaders"] = settings.response_headers,
        scheduler_class: Optional["SchedulerType"] = settings.scheduler_class,
        scheduler_tasks: Optional[Dict[str, str]] = settings.scheduler_tasks,
        scheduler_configurations: Optional[
            Dict[str, Union[str, Dict[str, str]]]
        ] = settings.scheduler_configurations,
        enable_scheduler: bool = settings.enable_scheduler,
        timezone: Optional[timezone] = settings.timezone,
        routes: Optional[List["APIGateHandler"]] = settings.routes,
        root_path: str = settings.root_path,
        middleware: Optional[Sequence["Middleware"]] = settings.middleware,
        exception_handlers: Optional["ExceptionHandlers"] = settings.exception_handlers,
        on_shutdown: Optional[List["LifeSpanHandler"]] = settings.on_shutdown,
        on_startup: Optional[List["LifeSpanHandler"]] = settings.on_startup,
        lifespan: Optional[Callable[["Esmerald"], "AsyncContextManager"]] = settings.lifespan,
        tags: Optional[List[str]] = settings.tags,
        include_in_schema: bool = settings.include_in_schema,
        deprecated: Optional[bool] = None,
        enable_openapi: bool = settings.enable_openapi,
        redirect_slashes: bool = settings.redirect_slashes,
    ) -> None:

        assert lifespan is None or (
            on_startup is None and on_shutdown is None
        ), "Use either 'lifespan' or 'on_startup'/'on_shutdown', not both."

        if allow_origins and cors_config:
            raise ImproperlyConfigured("It can be only allow_origins or cors_config but not both.")

        self._debug = debug
        self.title = title
        self.name = app_name
        self.description = description
        self.version = version
        self.summary = summary
        self.contact = contact
        self.terms_of_service = terms_of_service
        self.license_info = license
        self.servers = servers
        self.secret_key = secret_key
        self.allowed_hosts = allowed_hosts
        self.allow_origins = allow_origins
        self.dependencies = dependencies or {}
        self.permissions = permissions or []
        self.csrf_config = csrf_config
        self.cors_config = cors_config
        self.openapi_config = openapi_config
        self.openapi_schema: Optional["OpenAPI"] = None
        self.template_config = template_config
        self.static_files_config = static_files_config
        self.session_config = session_config
        self.response_class = response_class
        self.response_cookies = response_cookies
        self.response_headers = response_headers
        self.scheduler_class = scheduler_class
        self.scheduler_tasks = scheduler_tasks or {}
        self.scheduler_configurations = scheduler_configurations or {}
        self.enable_scheduler = enable_scheduler
        self.timezone = timezone
        self.tags = tags
        self.include_in_schema = include_in_schema
        self.middleware = middleware or []

        self.state = State()
        self.async_exit_config = settings.async_exit_config
        self.root_path = root_path
        self.parent: Optional[Union["ParentType", "Esmerald", "ChildEsmerald"]] = None
        self.on_shutdown = on_shutdown
        self.on_startup = on_startup
        self.security = security
        self.enable_openapi = enable_openapi
        self.redirect_slashes = redirect_slashes

        self.router: "Router" = Router(
            on_shutdown=on_shutdown,
            on_startup=on_startup,
            routes=routes,
            app=self,
            lifespan=lifespan,
            deprecated=deprecated,
            security=security,
            redirect_slashes=self.redirect_slashes,
        )

        self.exception_handlers = {} if exception_handlers is None else dict(exception_handlers)
        self.get_default_exception_handlers()

        self.user_middleware = self.build_user_middleware_stack()
        self.middleware_stack = self.build_middleware_stack()
        self.template_engine = self.get_template_engine(self.template_config)

        if self.static_files_config:
            for config in (
                self.static_files_config
                if isinstance(self.static_files_config, list)
                else [self.static_files_config]
            ):
                static_route = Include(path=config.path, app=config.to_app())
                self.router.validate_root_route_parent(static_route)
                self.router.routes.append(static_route)

        if self.enable_scheduler:
            self.scheduler = EsmeraldScheduler(
                app=self,
                scheduler_class=self.scheduler_class,
                tasks=self.scheduler_tasks,
                timezone=self.timezone,
                configurations=self.scheduler_configurations,
            )

        self.activate_openapi()

    def activate_openapi(self) -> None:
        if self.openapi_config and self.enable_openapi:
            self.openapi_schema = self.openapi_config.create_openapi_schema_model(self)
            gateway = gateways.Gateway(handler=self.openapi_config.openapi_apiview)
            self.router.add_apiview(value=gateway)

    def get_template_engine(
        self, template_config: "TemplateConfig"
    ) -> Optional["TemplateEngineProtocol"]:
        """
        Returns the template engine for the application.
        """
        if not template_config:
            return

        engine = template_config.engine(template_config.directory)
        return engine

    def add_route(
        self,
        path: str,
        handler: "HTTPHandler",
        router: Optional["Router"] = None,
        dependencies: Optional["Dependencies"] = None,
        exception_handlers: Optional["ExceptionHandlers"] = None,
        permissions: Optional[List["Permission"]] = None,
        middleware: Optional[List["Middleware"]] = None,
        name: Optional[str] = None,
        include_in_schema: bool = True,
    ) -> None:
        router = router or self.router
        route = router.add_route(
            path=path,
            handler=handler,
            dependencies=dependencies,
            exception_handlers=exception_handlers,
            permissions=permissions,
            middleware=middleware,
            name=name,
            include_in_schema=include_in_schema,
        )

        self.activate_openapi()
        return route

    def add_websocket_route(
        self,
        path: str,
        handler: "HTTPHandler",
        router: Optional["Router"] = None,
        dependencies: Optional["Dependencies"] = None,
        exception_handlers: Optional["ExceptionHandlers"] = None,
        permissions: Optional[List["Permission"]] = None,
        middleware: Optional[List["Middleware"]] = None,
        name: Optional[str] = None,
    ) -> None:
        router = router or self.router
        return router.add_websocket_route(
            path=path,
            handler=handler,
            dependencies=dependencies,
            exception_handlers=exception_handlers,
            permissions=permissions,
            middleware=middleware,
            name=name,
        )

    def add_router(self, router: "Router"):
        """
        Adds a router to the application.
        """
        for route in router.routes:
            if isinstance(route, Include):
                self.router.routes.append(
                    Include(
                        path=route.path,
                        app=route.app,
                        dependencies=route.dependencies,
                        exception_handlers=route.exception_handlers,
                        name=route.name,
                        middleware=route.middleware,
                        permissions=route.permissions,
                        routes=route.routes,
                        parent=self.router,
                    )
                )
                continue

            gateway = (
                gateways.Gateway
                if not isinstance(route.handler, WebSocketHandler)
                else gateways.WebSocketGateway
            )

            if self.on_startup:
                self.on_startup.extend(router.on_startup)
            if self.on_shutdown:
                self.on_shutdown.extend(router.on_shutdown)

            self.router.routes.append(
                gateway(
                    path=route.path,
                    dependencies=route.dependencies,
                    exception_handlers=route.exception_handlers,
                    name=route.name,
                    middleware=route.middleware,
                    permissions=route.permissions,
                    handler=route.handler,
                    parent=self.router,
                    is_from_router=True,
                )
            )

        self.activate_openapi()

    def get_default_exception_handlers(self) -> None:
        """
        Default exception handlers added to the application.
        Defaulting the HTTPException and ValidationError handlers to JSONResponse.

        The values can be overritten by calling the self.add_event_handler().

        Example:
            app = Esmerald()

            app.add_event_handler(HTTPException, my_new_handler)
            app.add_event_handler(ValidationError, my_new_422_handler)

        """
        self.exception_handlers.setdefault(
            ImproperlyConfigured, improperly_configured_exception_handler
        )
        self.exception_handlers.setdefault(
            ValidationErrorException, validation_error_exception_handler
        )

    def build_routes_middleware(
        self, route: "RouteParent", middlewares: Optional[List["Middleware"]] = None
    ):
        """
        Builds the middleware stack from the top to the bottom of the routes.
        We need to make sure the Esmerald/ChildEsmerald app.
        """
        if not middlewares:
            middlewares = []

        if isinstance(route, Include):
            app = getattr(route, "app", None)
            if app and isinstance(app, (Esmerald, ChildEsmerald)):
                return middlewares

            middlewares.extend(route.middleware)
            for _route in route.routes:
                middlewares = self.build_routes_middleware(_route, middlewares)

        if isinstance(route, (gateways.Gateway, gateways.WebSocketGateway)):
            middlewares.extend(route.middleware)
            if route.handler.middleware:
                middlewares.extend(route.handler.middleware)

        return middlewares

    def build_routes_exception_handlers(
        self,
        route: "RouteParent",
        exception_handlers: Optional["ExceptionHandlers"] = None,
    ):
        """
        Builds the exception handlers stack from the top to the bottom of the routes.
        """
        if not exception_handlers:
            exception_handlers = {}

        if isinstance(route, Include):
            exception_handlers.update(route.exception_handlers)
            app = getattr(route, "app", None)
            if app and isinstance(app, (Esmerald, ChildEsmerald)):
                return exception_handlers

            for _route in route.routes:
                exception_handlers = self.build_routes_exception_handlers(
                    _route, exception_handlers
                )

        if isinstance(route, (gateways.Gateway, gateways.WebSocketGateway)):
            exception_handlers.update(route.exception_handlers)
            if route.handler.exception_handlers:
                exception_handlers.update(route.handler.exception_handlers)

        return exception_handlers

    def build_user_middleware_stack(self) -> List["StarletteMiddleware"]:
        """
        Configures all the passed settings
        and wraps inside an exception handler.

        CORS, CSRF, TrustedHost and JWT are provided to the __init__ they will wapr the
        handler as well.

        It evaluates the middleware passed into the routes from bottom up
        """
        user_middleware = []
        handlers_middleware = []

        if self.allowed_hosts:
            user_middleware.append(
                StarletteMiddleware(TrustedHostMiddleware, allowed_hosts=self.allowed_hosts)
            )
        if self.cors_config:
            user_middleware.append(StarletteMiddleware(CORSMiddleware, **self.cors_config.dict()))
        if self.csrf_config:
            user_middleware.append(StarletteMiddleware(CSRFMiddleware, config=self.csrf_config))

        if self.session_config:
            user_middleware.append(
                StarletteMiddleware(SessionMiddleware, **self.session_config.dict())
            )

        handlers_middleware += self.router.middleware
        for route in self.routes or []:
            handlers_middleware.extend(self.build_routes_middleware(route))

        self.middleware += handlers_middleware

        for middleware in self.middleware or []:
            if isinstance(middleware, StarletteMiddleware):
                user_middleware.append(middleware)
            else:
                user_middleware.append(StarletteMiddleware(middleware))
        return user_middleware

    def build_middleware_stack(self) -> "ASGIApp":
        """
        Esmerald uses the [esmerald.protocols.MiddlewareProtocol] (interfaces) and therefore we
        wrap the StarletteMiddleware in a slighly different manner.

        Overriding the default build_middleware_stack will allow to control the initial
        middlewares that are loaded by Esmerald. All of these values can be updated.

        For each route, it will verify from top to bottom which exception handlers are being passed
        and called them accordingly.

        For APIViews, since it's a "wrapper", the handler will update the current list to contain
        both.
        """
        debug = self.debug
        error_handler = None
        exception_handlers = {}

        for key, value in self.exception_handlers.items():
            if key in (500, Exception):
                error_handler = value
            else:
                exception_handlers[key] = value

        for route in self.routes or []:
            exception_handlers.update(self.build_routes_exception_handlers(route))

        middleware = (
            [
                StarletteMiddleware(
                    EsmeraldAPIExceptionMiddleware,
                    exception_handlers=exception_handlers,
                    error_handler=error_handler,
                    debug=debug,
                ),
            ]
            + self.user_middleware
            + [
                StarletteMiddleware(
                    ExceptionMiddleware,
                    handlers=exception_handlers,
                    debug=debug,
                ),
                StarletteMiddleware(AsyncExitStackMiddleware, config=self.async_exit_config),
            ]
        )

        app: "ASGIApp" = self.router
        for cls, options in reversed(middleware):
            app = cls(app=app, **options)

        return app

    @property
    def settings(self) -> settings:
        """
        Returns the Esmerald settings object for easy access.
        """
        return settings

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        scope["app"] = self
        if scope["type"] == "lifespan":
            await self.router.lifespan(scope, receive, send)
            return
        if self.root_path:
            scope["root_path"] = self.root_path
        scope["state"] = {}
        await super().__call__(scope, receive, send)

    def mount(self, path: str, app: ASGIApp, name: Optional[str] = None) -> None:
        raise ImproperlyConfigured("`mount` is not supported by Esmerald. Use Include() instead.")

    def host(self, host: str, app: ASGIApp, name: Optional[str] = None) -> None:
        raise ImproperlyConfigured(
            "`host` is not supported by Esmerald. Use Starlette Host() instead."
        )

    def route(
        self,
        path: str,
        methods: Optional[List[str]] = None,
        name: Optional[str] = None,
        include_in_schema: bool = True,
    ) -> Callable:
        raise ImproperlyConfigured("`route` is not valid. Use Gateway instead.")

    def websocket_route(self, path: str, name: Optional[str] = None) -> Callable:
        raise ImproperlyConfigured("`websocket_route` is not valid. Use WebSocketGateway instead.")


class ChildEsmerald(Esmerald):
    """
    A simple visual representation of an Esmerald application, exactly like Esmerald
    but for organisation purposes it might be preferred to use this class.
    """

    ...



from typing import Any, Dict, Optional, Union

from esmerald.enums import EncodingType, ParamType
from esmerald.utils.constants import IS_DEPENDENCY, SKIP_VALIDATION
from pydantic.fields import FieldInfo, Undefined


class Param(FieldInfo):
    in_: ParamType

    def __init__(
        self,
        default: Any = Undefined,
        *,
        alias: Optional[str] = None,
        value_type: Any = Undefined,
        header: Optional[str] = None,
        cookie: Optional[str] = None,
        query: Optional[str] = None,
        path: Optional[str] = None,
        content_encoding: Optional[str] = None,
        required: bool = True,
        title: Optional[str] = None,
        description: Optional[str] = None,
        const: Optional[bool] = None,
        gt: Optional[float] = None,
        ge: Optional[float] = None,
        lt: Optional[float] = None,
        le: Optional[float] = None,
        multiple_of: Optional[float] = None,
        min_items: Optional[int] = None,
        max_items: Optional[int] = None,
        min_length: Optional[int] = None,
        max_length: Optional[int] = None,
        regex: Optional[str] = None,
        example: Any = Undefined,
        examples: Optional[Dict[str, Any]] = None,
        deprecated: Optional[bool] = None,
        include_in_schema: bool = True,
    ) -> Any:
        self.deprecated = deprecated
        self.example = example
        self.examples = examples
        self.include_in_schema = include_in_schema

        extra: Dict[Union[str, int], Any] = {}
        extra.update(header=header)
        extra.update(cookie=cookie)
        extra.update(query=query)
        extra.update(path=path)
        extra.update(required=required)
        extra.update(content_encoding=content_encoding)
        extra.update(value_type=value_type)
        extra.update(example=self.example)
        extra.update(examples=self.examples)
        extra.update(deprecated=self.deprecated)
        extra.update(include_in_schema=self.include_in_schema)

        super().__init__(
            default=default,
            alias=alias,
            title=title,
            description=description,
            const=const,
            gt=gt,
            ge=ge,
            lt=lt,
            le=le,
            multiple_of=multiple_of,
            min_items=min_items,
            max_items=max_items,
            min_length=min_length,
            max_length=max_length,
            regex=regex,
            **extra,
        )


class Header(Param):
    in_ = ParamType.HEADER

    def __init__(
        self,
        *,
        default: Any = Undefined,
        value: Optional[str] = None,
        alias: Optional[str] = None,
        value_type: Any = Undefined,
        content_encoding: Optional[str] = None,
        required: bool = True,
        title: Optional[str] = None,
        description: Optional[str] = None,
        const: Optional[bool] = None,
        gt: Optional[float] = None,
        ge: Optional[float] = None,
        lt: Optional[float] = None,
        le: Optional[float] = None,
        multiple_of: Optional[float] = None,
        min_items: Optional[int] = None,
        max_items: Optional[int] = None,
        min_length: Optional[int] = None,
        max_length: Optional[int] = None,
        regex: Optional[str] = None,
        example: Any = Undefined,
        examples: Optional[Dict[str, Any]] = None,
        deprecated: Optional[bool] = None,
        include_in_schema: bool = True,
    ) -> Any:
        self.in_ = self.in_
        super().__init__(
            default=default,
            header=value,
            alias=alias,
            title=title,
            description=description,
            const=const,
            gt=gt,
            ge=ge,
            lt=lt,
            le=le,
            multiple_of=multiple_of,
            min_items=min_items,
            max_items=max_items,
            min_length=min_length,
            max_length=max_length,
            regex=regex,
            required=required,
            content_encoding=content_encoding,
            value_type=value_type,
            example=example,
            examples=examples,
            deprecated=deprecated,
            include_in_schema=include_in_schema,
        )


class Cookie(Param):
    in_ = ParamType.COOKIE

    def __init__(
        self,
        default: Any = Undefined,
        *,
        value_type: Any = Undefined,
        value: Optional[str] = None,
        alias: Optional[str] = None,
        content_encoding: Optional[str] = None,
        required: bool = True,
        title: Optional[str] = None,
        description: Optional[str] = None,
        const: Optional[bool] = None,
        gt: Optional[float] = None,
        ge: Optional[float] = None,
        lt: Optional[float] = None,
        le: Optional[float] = None,
        multiple_of: Optional[float] = None,
        min_items: Optional[int] = None,
        max_items: Optional[int] = None,
        min_length: Optional[int] = None,
        max_length: Optional[int] = None,
        regex: Optional[str] = None,
        example: Any = Undefined,
        examples: Optional[Dict[str, Any]] = None,
        deprecated: Optional[bool] = None,
        include_in_schema: bool = True,
    ) -> Any:
        self.in_ = self.in_
        super().__init__(
            default=default,
            cookie=value,
            alias=alias,
            title=title,
            description=description,
            const=const,
            gt=gt,
            ge=ge,
            lt=lt,
            le=le,
            multiple_of=multiple_of,
            min_items=min_items,
            max_items=max_items,
            min_length=min_length,
            max_length=max_length,
            regex=regex,
            required=required,
            content_encoding=content_encoding,
            value_type=value_type,
            example=example,
            examples=examples,
            deprecated=deprecated,
            include_in_schema=include_in_schema,
        )


class Query(Param):
    in_ = ParamType.QUERY

    def __init__(
        self,
        default: Any = Undefined,
        *,
        value_type: Any = Undefined,
        value: Optional[str] = None,
        alias: Optional[str] = None,
        content_encoding: Optional[str] = None,
        required: bool = True,
        title: Optional[str] = None,
        description: Optional[str] = None,
        const: Optional[bool] = None,
        gt: Optional[float] = None,
        ge: Optional[float] = None,
        lt: Optional[float] = None,
        le: Optional[float] = None,
        multiple_of: Optional[float] = None,
        min_items: Optional[int] = None,
        max_items: Optional[int] = None,
        min_length: Optional[int] = None,
        max_length: Optional[int] = None,
        regex: Optional[str] = None,
        example: Any = Undefined,
        examples: Optional[Dict[str, Any]] = None,
        deprecated: Optional[bool] = None,
        include_in_schema: bool = True,
    ) -> Any:
        self.in_ = self.in_
        super().__init__(
            default=default,
            query=value,
            alias=alias,
            title=title,
            description=description,
            const=const,
            gt=gt,
            ge=ge,
            lt=lt,
            le=le,
            multiple_of=multiple_of,
            min_items=min_items,
            max_items=max_items,
            min_length=min_length,
            max_length=max_length,
            regex=regex,
            required=required,
            content_encoding=content_encoding,
            value_type=value_type,
            example=example,
            examples=examples,
            deprecated=deprecated,
            include_in_schema=include_in_schema,
        )


class Path(Param):
    in_ = ParamType.PATH

    def __init__(
        self,
        default: Any = Undefined,
        *,
        value_type: Any = Undefined,
        content_encoding: Optional[str] = None,
        required: bool = True,
        title: Optional[str] = None,
        description: Optional[str] = None,
        const: Optional[bool] = None,
        gt: Optional[float] = None,
        ge: Optional[float] = None,
        lt: Optional[float] = None,
        le: Optional[float] = None,
        min_length: Optional[int] = None,
        max_length: Optional[int] = None,
        regex: Optional[str] = None,
        example: Any = Undefined,
        examples: Optional[Dict[str, Any]] = None,
        deprecated: Optional[bool] = None,
        include_in_schema: bool = True,
    ) -> Any:
        self.in_ = self.in_
        super().__init__(
            default=default,
            title=title,
            description=description,
            const=const,
            gt=gt,
            ge=ge,
            lt=lt,
            le=le,
            min_length=min_length,
            max_length=max_length,
            regex=regex,
            required=required,
            content_encoding=content_encoding,
            value_type=value_type,
            example=example,
            examples=examples,
            deprecated=deprecated,
            include_in_schema=include_in_schema,
        )


class Body(FieldInfo):
    def __init__(
        self,
        *,
        default: Any = Undefined,
        media_type: Union[str, EncodingType] = EncodingType.JSON,
        content_encoding: Optional[str] = None,
        title: Optional[str] = None,
        alias: Optional[str] = None,
        description: Optional[str] = None,
        const: Optional[bool] = None,
        gt: Optional[float] = None,
        ge: Optional[float] = None,
        lt: Optional[float] = None,
        le: Optional[float] = None,
        multiple_of: Optional[float] = None,
        min_items: Optional[int] = None,
        max_items: Optional[int] = None,
        min_length: Optional[int] = None,
        max_length: Optional[int] = None,
        regex: Optional[str] = None,
        example: Any = Undefined,
        examples: Optional[Dict[str, Any]] = None,
    ) -> Any:
        extra: Dict[str, Any] = {}
        self.media_type = media_type
        self.content_encoding = content_encoding
        self.example = example
        self.examples = examples
        extra.update(media_type=self.media_type)
        extra.update(content_encoding=self.content_encoding)
        super().__init__(
            default=default,
            title=title,
            alias=alias,
            description=description,
            const=const,
            gt=gt,
            ge=ge,
            lt=lt,
            le=le,
            multiple_of=multiple_of,
            min_items=min_items,
            max_items=max_items,
            min_length=min_length,
            max_length=max_length,
            regex=regex,
            **extra,
        )


class Injects(FieldInfo):
    """
    Creates a FieldInfo class with extra parameters.
    This is used for dependencies and to inject them.

    Example:

        @get(dependencies={"value": Inject(lambda: 13)})
        def myview(value: Injects()):
            return {"value": value}
    """

    def __init__(
        self,
        default: Any = Undefined,
        skip_validation: bool = False,
    ) -> None:
        extra: Dict[str, Any] = {
            IS_DEPENDENCY: True,
            SKIP_VALIDATION: skip_validation,
        }
        super().__init__(default, **extra)



from http import HTTPStatus
from typing import Any, Dict, Optional, Type, Union

from pydantic import BaseModel, create_model
from starlette import status
from starlette.exceptions import HTTPException as StarletteHTTPException
from starlette.exceptions import WebSocketException

RequestErrorModel: Type[BaseModel] = create_model("Request")
WebSocketErrorModel: Type[BaseModel] = create_model("WebSocket")


class EsmeraldAPIException(Exception):
    def __init__(self, *args: Any, detail: str = ""):
        self.detail = detail
        super().__init__(*(str(arg) for arg in args if arg), detail)

    def __repr__(self) -> str:
        if self.detail:
            return f"{self.__class__.__name__} - {self.detail}"
        return self.__class__.__name__

    def __str__(self) -> str:
        return "".join(self.args).strip()


class HTTPException(StarletteHTTPException, EsmeraldAPIException):
    """
    Implementation of a unique exception allowing to override
    the details.
    """

    status_code = status.HTTP_500_INTERNAL_SERVER_ERROR

    def __init__(
        self,
        *args: Any,
        detail: Optional[str] = None,
        status_code: Optional[int] = None,
        headers: Optional[Dict[str, Any]] = None,
        **extra,
    ):
        detail = detail or getattr(self, "detail", None)
        status_code = status_code or getattr(self, "status_code", None)
        if not detail:
            detail = args[0] if args else HTTPStatus(status_code or self.status_code).phrase
            args = args[1:]
        super().__init__(status_code=status_code, detail=detail, headers=headers)
        self.detail = detail
        self.headers = headers
        self.args = (f"{self.status_code}: {self.detail}", *args)
        self.extra = extra

    def __repr__(self) -> str:
        return f"<{self.status_code}: {self.__class__.__name__} />"


class EsmeraldError(RuntimeError):
    """
    Generic exception.
    """

    ...


class ImproperlyConfigured(HTTPException, ValueError):
    ...


class NotAuthenticated(HTTPException):
    status_code = status.HTTP_401_UNAUTHORIZED
    detail = "Authentication credentials were not provided."


class PermissionDenied(HTTPException):
    status_code = status.HTTP_403_FORBIDDEN
    detail = "You do not have permission to perform this action."


class ValidationErrorException(HTTPException, ValueError):
    status_code = status.HTTP_400_BAD_REQUEST


class NotAuthorized(HTTPException):
    status_code = status.HTTP_401_UNAUTHORIZED
    detail = "You do not have authorization to perform this action."


class NotFound(HTTPException, ValueError):
    detail = "The resource cannot be found."
    status_code = status.HTTP_404_NOT_FOUND


class MethodNotAllowed(HTTPException):
    status_code = status.HTTP_405_METHOD_NOT_ALLOWED


class InternalServerError(HTTPException):
    status_code = status.HTTP_500_INTERNAL_SERVER_ERROR


class ServiceUnavailable(HTTPException):
    status_code = status.HTTP_503_SERVICE_UNAVAILABLE


class TemplateNotFound(InternalServerError):
    def __init__(self, *args: Any, template_name: str):
        """Template could not be found."""
        super().__init__(*args, detail=f"Template {template_name} not found.")


class MissingDependency(EsmeraldAPIException, ImportError):
    ...


class OpenAPIError(ValueError):
    ...


class WebSocketException(WebSocketException):
    ...


ExceptionErrorMap = Union[
    HTTPException,
    NotAuthenticated,
    PermissionDenied,
    NotAuthorized,
    NotFound,
    MethodNotAllowed,
    InternalServerError,
    ServiceUnavailable,
]



from typing import Union

from pydantic.error_wrappers import ValidationError
from starlette import status
from starlette.exceptions import HTTPException as StarletteHTTPException
from starlette.requests import Request
from starlette.responses import Response as StarletteResponse

from esmerald.enums import MediaType
from esmerald.exceptions import ExceptionErrorMap, HTTPException, ImproperlyConfigured
from esmerald.responses import ORJSONResponse, Response


async def http_exception_handler(
    request: Request, exc: Union[HTTPException, StarletteHTTPException]
) -> Union[ORJSONResponse, Response]:
    """
    Default exception handler for StarletteHTTPException and Esmerald HTTPException.
    """
    extra = getattr(exc, "extra", None)
    headers = getattr(exc, "headers", None)

    if exc.status_code in {204, 304}:
        return ORJSONResponse(status_code=exc.status_code, headers=headers)

    if headers and not extra:
        return ORJSONResponse({"detail": exc.detail}, status_code=exc.status_code, headers=headers)
    elif headers and extra:
        return ORJSONResponse(
            {"detail": exc.detail, "extra": extra},
            status_code=exc.status_code,
            headers=headers,
        )
    elif not headers and extra:
        return ORJSONResponse({"detail": exc.detail, "extra": extra}, status_code=exc.status_code)
    else:
        return ORJSONResponse({"detail": exc.detail}, status_code=exc.status_code)


async def validation_error_exception_handler(
    request: Request, exc: ValidationError
) -> ORJSONResponse:
    extra = getattr(exc, "extra", None)
    status_code = status.HTTP_400_BAD_REQUEST

    if extra:
        return ORJSONResponse(
            {"detail": exc.detail, "errors": exc.extra.get("extra", {})},
            status_code=status_code,
        )
    else:
        return ORJSONResponse(
            {"detail": exc.detail},
            status_code=status_code,
        )


async def http_error_handler(_: Request, exc: ExceptionErrorMap) -> ORJSONResponse:
    return ORJSONResponse({"detail": exc.detail}, status_code=exc.status_code)


async def improperly_configured_exception_handler(
    request: Request, exc: ImproperlyConfigured
) -> StarletteResponse:
    status_code = (
        exc.status_code
        if isinstance(exc, StarletteHTTPException)
        else status.HTTP_500_INTERNAL_SERVER_ERROR
    )
    if not status_code:
        status_code = status.HTTP_500_INTERNAL_SERVER_ERROR

    content = {"detail": exc.detail}
    if exc.extra:
        content.update({"extra": exc.extra})
    headers = exc.headers if isinstance(exc, (HTTPException, StarletteHTTPException)) else None

    return Response(
        media_type=MediaType.JSON,
        content=content,
        status_code=status_code,
        headers=headers,
    )



from enum import Enum
from typing import TYPE_CHECKING, Dict, List, Optional, Type, Union

from esmerald.enums import HttpMethod, MediaType
from esmerald.exceptions import HTTPException, ImproperlyConfigured
from esmerald.openapi.datastructures import ResponseSpecification
from esmerald.permissions.types import Permission
from esmerald.routing.router import HTTPHandler, WebSocketHandler
from esmerald.types import (
    BackgroundTaskType,
    Dependencies,
    ExceptionHandler,
    ExceptionHandlers,
    HTTPMethod,
    Middleware,
    ResponseCookies,
    ResponseHeaders,
    ResponseType,
)
from esmerald.utils.constants import AVAILABLE_METHODS
from starlette import status

if TYPE_CHECKING:
    from esmerald.exceptions import HTTPException
    from openapi_schemas_pydantic.v3_1_0 import SecurityRequirement


class get(HTTPHandler):
    def __init__(
        self,
        path: Optional[str] = None,
        *,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        status_code: Optional[int] = status.HTTP_200_OK,
        content_encoding: Optional[str] = None,
        content_media_type: Optional[str] = None,
        include_in_schema: bool = True,
        background: Optional["BackgroundTaskType"] = None,
        dependencies: Optional["Dependencies"] = None,
        exception_handlers: Optional[ExceptionHandlers] = None,
        permissions: Optional[List[Permission]] = None,
        middleware: Optional[List[Middleware]] = None,
        media_type: Union[MediaType, str] = MediaType.JSON,
        response_class: Optional[ResponseType] = None,
        response_cookies: Optional[ResponseCookies] = None,
        response_headers: Optional[ResponseHeaders] = None,
        tags: Optional[List[Union[str, Enum]]] = None,
        deprecated: Optional[bool] = None,
        security: Optional[List["SecurityRequirement"]] = None,
        operation_id: Optional[str] = None,
        raise_exceptions: Optional[List[Type["HTTPException"]]] = None,
        response_description: Optional[str] = "Successful response",
        responses: Optional[Dict[int, ResponseSpecification]] = None,
    ) -> None:
        super().__init__(
            path=path,
            methods=[HttpMethod.GET],
            summary=summary,
            description=description,
            status_code=status_code,
            content_encoding=content_encoding,
            content_media_type=content_media_type,
            include_in_schema=include_in_schema,
            background=background,
            dependencies=dependencies,
            exception_handlers=exception_handlers,
            permissions=permissions,
            middleware=middleware,
            media_type=media_type,
            response_class=response_class,
            response_cookies=response_cookies,
            response_headers=response_headers,
            tags=tags,
            deprecated=deprecated,
            security=security,
            operation_id=operation_id,
            raise_exceptions=raise_exceptions,
            response_description=response_description,
            responses=responses,
        )


class post(HTTPHandler):
    def __init__(
        self,
        path: Optional[str] = None,
        *,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        status_code: Optional[int] = status.HTTP_201_CREATED,
        content_encoding: Optional[str] = None,
        content_media_type: Optional[str] = None,
        include_in_schema: bool = True,
        background: Optional["BackgroundTaskType"] = None,
        dependencies: Optional["Dependencies"] = None,
        exception_handlers: Optional[ExceptionHandlers] = None,
        permissions: Optional[List[Permission]] = None,
        middleware: Optional[List[Middleware]] = None,
        media_type: Union[MediaType, str] = MediaType.JSON,
        response_class: Optional[ResponseType] = None,
        response_cookies: Optional[ResponseCookies] = None,
        response_headers: Optional[ResponseHeaders] = None,
        tags: Optional[List[Union[str, Enum]]] = None,
        deprecated: Optional[bool] = None,
        security: Optional[List["SecurityRequirement"]] = None,
        operation_id: Optional[str] = None,
        raise_exceptions: Optional[List[Type["HTTPException"]]] = None,
        response_description: Optional[str] = "Successful response",
        responses: Optional[Dict[int, ResponseSpecification]] = None,
    ) -> None:
        super().__init__(
            path=path,
            status_code=status_code,
            content_encoding=content_encoding,
            content_media_type=content_media_type,
            summary=summary,
            description=description,
            methods=[HttpMethod.POST],
            include_in_schema=include_in_schema,
            background=background,
            dependencies=dependencies,
            exception_handlers=exception_handlers,
            permissions=permissions,
            middleware=middleware,
            media_type=media_type,
            response_class=response_class,
            response_cookies=response_cookies,
            response_headers=response_headers,
            tags=tags,
            deprecated=deprecated,
            security=security,
            operation_id=operation_id,
            raise_exceptions=raise_exceptions,
            response_description=response_description,
            responses=responses,
        )


class put(HTTPHandler):
    def __init__(
        self,
        path: Optional[str] = None,
        *,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        status_code: Optional[int] = status.HTTP_200_OK,
        content_encoding: Optional[str] = None,
        content_media_type: Optional[str] = None,
        include_in_schema: bool = True,
        background: Optional["BackgroundTaskType"] = None,
        dependencies: Optional["Dependencies"] = None,
        exception_handlers: Optional[ExceptionHandlers] = None,
        permissions: Optional[List[Permission]] = None,
        middleware: Optional[List[Middleware]] = None,
        media_type: Union[MediaType, str] = MediaType.JSON,
        response_class: Optional[ResponseType] = None,
        response_cookies: Optional[ResponseCookies] = None,
        response_headers: Optional[ResponseHeaders] = None,
        tags: Optional[List[Union[str, Enum]]] = None,
        deprecated: Optional[bool] = None,
        security: Optional[List["SecurityRequirement"]] = None,
        operation_id: Optional[str] = None,
        raise_exceptions: Optional[List[Type["HTTPException"]]] = None,
        response_description: Optional[str] = "Successful response",
        responses: Optional[Dict[int, ResponseSpecification]] = None,
    ) -> None:
        super().__init__(
            path=path,
            methods=[HttpMethod.PUT],
            summary=summary,
            description=description,
            status_code=status_code,
            content_encoding=content_encoding,
            content_media_type=content_media_type,
            include_in_schema=include_in_schema,
            background=background,
            dependencies=dependencies,
            exception_handlers=exception_handlers,
            permissions=permissions,
            middleware=middleware,
            media_type=media_type,
            response_class=response_class,
            response_cookies=response_cookies,
            response_headers=response_headers,
            tags=tags,
            deprecated=deprecated,
            security=security,
            operation_id=operation_id,
            raise_exceptions=raise_exceptions,
            response_description=response_description,
            responses=responses,
        )


class patch(HTTPHandler):
    def __init__(
        self,
        path: Optional[str] = None,
        *,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        status_code: Optional[int] = status.HTTP_200_OK,
        content_encoding: Optional[str] = None,
        content_media_type: Optional[str] = None,
        include_in_schema: bool = True,
        background: Optional["BackgroundTaskType"] = None,
        dependencies: Optional["Dependencies"] = None,
        exception_handlers: Optional[ExceptionHandlers] = None,
        permissions: Optional[List[Permission]] = None,
        middleware: Optional[List[Middleware]] = None,
        media_type: Union[MediaType, str] = MediaType.JSON,
        response_class: Optional[ResponseType] = None,
        response_cookies: Optional[ResponseCookies] = None,
        response_headers: Optional[ResponseHeaders] = None,
        tags: Optional[List[Union[str, Enum]]] = None,
        deprecated: Optional[bool] = None,
        security: Optional[List["SecurityRequirement"]] = None,
        operation_id: Optional[str] = None,
        raise_exceptions: Optional[List[Type["HTTPException"]]] = None,
        response_description: Optional[str] = "Successful response",
        responses: Optional[Dict[int, ResponseSpecification]] = None,
    ) -> None:
        super().__init__(
            path=path,
            methods=[HttpMethod.PATCH],
            summary=summary,
            description=description,
            status_code=status_code,
            content_encoding=content_encoding,
            content_media_type=content_media_type,
            include_in_schema=include_in_schema,
            background=background,
            dependencies=dependencies,
            exception_handlers=exception_handlers,
            permissions=permissions,
            middleware=middleware,
            media_type=media_type,
            response_class=response_class,
            response_cookies=response_cookies,
            response_headers=response_headers,
            tags=tags,
            deprecated=deprecated,
            security=security,
            operation_id=operation_id,
            raise_exceptions=raise_exceptions,
            response_description=response_description,
            responses=responses,
        )


class delete(HTTPHandler):
    def __init__(
        self,
        path: Optional[str] = None,
        *,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        status_code: Optional[int] = status.HTTP_204_NO_CONTENT,
        content_encoding: Optional[str] = None,
        content_media_type: Optional[str] = None,
        include_in_schema: bool = True,
        background: Optional["BackgroundTaskType"] = None,
        dependencies: Optional["Dependencies"] = None,
        exception_handlers: Optional[ExceptionHandlers] = None,
        permissions: Optional[List[Permission]] = None,
        middleware: Optional[List[Middleware]] = None,
        media_type: Union[MediaType, str] = MediaType.JSON,
        response_class: Optional[ResponseType] = None,
        response_cookies: Optional[ResponseCookies] = None,
        response_headers: Optional[ResponseHeaders] = None,
        tags: Optional[List[Union[str, Enum]]] = None,
        deprecated: Optional[bool] = None,
        security: Optional[List["SecurityRequirement"]] = None,
        operation_id: Optional[str] = None,
        raise_exceptions: Optional[List[Type["HTTPException"]]] = None,
        response_description: Optional[str] = "Successful response",
        responses: Optional[Dict[int, ResponseSpecification]] = None,
    ) -> None:
        super().__init__(
            path=path,
            methods=[HttpMethod.DELETE],
            summary=summary,
            description=description,
            status_code=status_code,
            content_encoding=content_encoding,
            content_media_type=content_media_type,
            include_in_schema=include_in_schema,
            background=background,
            dependencies=dependencies,
            exception_handlers=exception_handlers,
            permissions=permissions,
            middleware=middleware,
            media_type=media_type,
            response_class=response_class,
            response_cookies=response_cookies,
            response_headers=response_headers,
            tags=tags,
            deprecated=deprecated,
            security=security,
            operation_id=operation_id,
            raise_exceptions=raise_exceptions,
            response_description=response_description,
            responses=responses,
        )


class route(HTTPHandler):
    def __init__(
        self,
        path: Optional[str] = None,
        *,
        methods: List["HTTPMethod"] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        status_code: Optional[int] = status.HTTP_200_OK,
        content_encoding: Optional[str] = None,
        content_media_type: Optional[str] = None,
        include_in_schema: bool = True,
        background: Optional["BackgroundTaskType"] = None,
        dependencies: Optional["Dependencies"] = None,
        exception_handlers: Optional[ExceptionHandlers] = None,
        permissions: Optional[List[Permission]] = None,
        middleware: Optional[List[Middleware]] = None,
        media_type: Union[MediaType, str] = MediaType.JSON,
        response_class: Optional[ResponseType] = None,
        response_cookies: Optional[ResponseCookies] = None,
        response_headers: Optional[ResponseHeaders] = None,
        tags: Optional[List[Union[str, Enum]]] = None,
        deprecated: Optional[bool] = None,
        security: Optional[List["SecurityRequirement"]] = None,
        operation_id: Optional[str] = None,
        raise_exceptions: Optional[List[Type["HTTPException"]]] = None,
        response_description: Optional[str] = "Successful response",
        responses: Optional[Dict[int, ResponseSpecification]] = None,
    ) -> None:
        if not methods or not isinstance(methods, list):
            raise ImproperlyConfigured(
                "http handler demands `methods` to be declared. "
                "An example would be: @route(methods=['GET', 'PUT'])."
            )

        for method in methods:
            if method.upper() not in AVAILABLE_METHODS:
                raise ImproperlyConfigured(
                    f"Invalid method {method}. "
                    "An example would be: @route(methods=['GET', 'PUT'])."
                )

        methods = [method.upper() for method in methods]
        if not status_code:
            status_code = status.HTTP_200_OK

        super().__init__(
            path=path,
            methods=methods,
            summary=summary,
            description=description,
            status_code=status_code,
            content_encoding=content_encoding,
            content_media_type=content_media_type,
            include_in_schema=include_in_schema,
            background=background,
            dependencies=dependencies,
            exception_handlers=exception_handlers,
            permissions=permissions,
            middleware=middleware,
            media_type=media_type,
            response_class=response_class,
            response_cookies=response_cookies,
            response_headers=response_headers,
            tags=tags,
            deprecated=deprecated,
            security=security,
            operation_id=operation_id,
            raise_exceptions=raise_exceptions,
            response_description=response_description,
            responses=responses,
        )


class websocket(WebSocketHandler):
    def __init__(
        self,
        path: Union[Optional[str], Optional[List[str]]] = None,
        *,
        dependencies: Optional["Dependencies"] = None,
        exception_handlers: Optional[Dict[Union[int, Type[Exception]], "ExceptionHandler"]] = None,
        permissions: Optional[List["Permission"]] = None,
        middleware: Optional[List["Middleware"]] = None,
    ):
        super().__init__(
            path=path,
            dependencies=dependencies,
            exception_handlers=exception_handlers,
            permissions=permissions,
            middleware=middleware,
        )



import inspect
from typing import Any

from esmerald.transformers.types import ConstrainedField
from pydantic import (
    ConstrainedBytes,
    ConstrainedDate,
    ConstrainedDecimal,
    ConstrainedFloat,
    ConstrainedFrozenSet,
    ConstrainedInt,
    ConstrainedList,
    ConstrainedSet,
    ConstrainedStr,
)


def is_pydantic_constrained_field(value: Any) -> ConstrainedField:
    return inspect.isclass(value) and any(
        issubclass(value, _type)
        for _type in (
            ConstrainedBytes,
            ConstrainedDate,
            ConstrainedDecimal,
            ConstrainedFloat,
            ConstrainedFrozenSet,
            ConstrainedInt,
            ConstrainedList,
            ConstrainedSet,
            ConstrainedStr,
        )
    )



from dataclasses import is_dataclass
from decimal import Decimal
from enum import Enum, EnumMeta
from typing import Any, List, Optional, Type, Union

from esmerald.datastructures import UploadFile
from esmerald.openapi.enums import OpenAPIType
from esmerald.openapi.utils import get_openapi_type_for_complex_type
from esmerald.utils.model import convert_dataclass_to_model, create_parsed_model_field
from openapi_schemas_pydantic.utils.constants import (
    EXTRA_TO_OPENAPI_PROPERTY_MAP,
    PYDANTIC_TO_OPENAPI_PROPERTY_MAP,
    TYPE_MAP,
)
from openapi_schemas_pydantic.utils.utils import OpenAPI310PydanticSchema
from openapi_schemas_pydantic.v3_1_0.example import Example
from openapi_schemas_pydantic.v3_1_0.schema import Schema
from pydantic import (
    BaseModel,
    ConstrainedBytes,
    ConstrainedDecimal,
    ConstrainedFloat,
    ConstrainedInt,
    ConstrainedList,
    ConstrainedSet,
    ConstrainedStr,
)
from pydantic.fields import FieldInfo, ModelField, Undefined
from pydantic_factories import ModelFactory
from pydantic_factories.exceptions import ParameterError
from pydantic_factories.utils import is_optional, is_pydantic_model, is_union


def clean_values_from_example(value: Any):
    if isinstance(value, (Decimal, float)):
        value = round(float(value), 3)

    if isinstance(value, Enum):
        value = value.value

    if is_dataclass(value):
        value = convert_dataclass_to_model(value)

    if isinstance(value, BaseModel):
        value = value.dict()

    if isinstance(value, (list, set)):
        value = [clean_values_from_example(v) for v in value]

    if isinstance(value, dict):
        for k, v in value.items():
            value[k] = clean_values_from_example(v)

    return value


class ExampleFactory(ModelFactory[BaseModel]):

    __model__ = BaseModel
    __allow_none_optionals__: bool = False


def create_numerical_constrained_field_schema(
    field_type: Union[Type[ConstrainedFloat], Type[ConstrainedInt], Type[ConstrainedDecimal]]
) -> Schema:

    schema = Schema(
        type=OpenAPIType.INTEGER if issubclass(field_type, int) else OpenAPIType.NUMBER
    )
    if field_type.le is not None:
        schema.maximum = float(field_type.le)
    if field_type.lt is not None:
        schema.exclusiveMaximum = float(field_type.lt)
    if field_type.ge is not None:
        schema.minimum = float(field_type.ge)
    if field_type.gt is not None:
        schema.exclusiveMinimum = float(field_type.gt)
    if field_type.multiple_of is not None:
        schema.multipleOf = float(field_type.multiple_of)
    return schema


def create_string_constrained_field_schema(
    field_type: Union[Type[ConstrainedStr], Type[ConstrainedBytes]]
) -> Schema:
    schema = Schema(type=OpenAPIType.STRING)
    if field_type.min_length:
        schema.minLength = field_type.min_length
    if field_type.max_length:
        schema.maxLength = field_type.max_length
    if issubclass(field_type, ConstrainedStr) and field_type.regex is not None:
        schema.pattern = field_type.regex.pattern
    if field_type.to_lower:
        schema.description = "must be in lower case"
    return schema


def create_collection_constrained_field_schema(
    field_type: Union[Type[ConstrainedList], Type[ConstrainedSet]],
    sub_fields: Optional[List[ModelField]],
) -> Schema:
    """Create Schema from Constrained List/Set field."""
    schema = Schema(type=OpenAPIType.ARRAY)
    if field_type.min_items:
        schema.minItems = field_type.min_items
    if field_type.max_items:
        schema.maxItems = field_type.max_items
    if issubclass(field_type, ConstrainedSet):
        schema.uniqueItems = True
    if sub_fields:
        items = [create_schema(field=sub_field, create_examples=False) for sub_field in sub_fields]
        if len(items) > 1:
            schema.items = Schema(oneOf=items)  # type: ignore[arg-type]
        else:
            schema.items = items[0]
    else:
        parsed_model_field = create_parsed_model_field(field_type.item_type)
        schema.items = create_schema(field=parsed_model_field, create_examples=False)
    return schema


def create_constrained_field_schema(
    field_type: Union[
        Type[ConstrainedSet],
        Type[ConstrainedList],
        Type[ConstrainedStr],
        Type[ConstrainedBytes],
        Type[ConstrainedFloat],
        Type[ConstrainedInt],
        Type[ConstrainedDecimal],
    ],
    sub_fields: Optional[List[ModelField]],
) -> Schema:
    if issubclass(field_type, (ConstrainedFloat, ConstrainedInt, ConstrainedDecimal)):
        return create_numerical_constrained_field_schema(field_type=field_type)
    if issubclass(field_type, (ConstrainedStr, ConstrainedBytes)):
        return create_string_constrained_field_schema(field_type=field_type)
    return create_collection_constrained_field_schema(field_type=field_type, sub_fields=sub_fields)


def update_schema_field_info(schema: Schema, field_info: FieldInfo) -> Schema:
    if (
        field_info.const
        and field_info.default not in [None, ..., Undefined]
        and schema.const is None
    ):
        schema.const = field_info.default
    for pydantic_key, schema_key in PYDANTIC_TO_OPENAPI_PROPERTY_MAP.items():
        value = getattr(field_info, pydantic_key)
        if value not in [None, ..., Undefined]:
            setattr(schema, schema_key, value)

    for extra_key, schema_key in EXTRA_TO_OPENAPI_PROPERTY_MAP.items():
        if extra_key in field_info.extra:
            value = field_info.extra[extra_key]
            if value not in [None, ..., Undefined]:
                setattr(schema, schema_key, value)
    return schema


def get_schema_for_field_type(field: ModelField) -> Schema:
    field_type = field.outer_type_
    if field_type in TYPE_MAP:
        return TYPE_MAP[field_type].copy()
    if is_pydantic_model(field_type):
        return OpenAPI310PydanticSchema(schema_class=field_type)
    if is_dataclass(field_type):
        return OpenAPI310PydanticSchema(schema_class=convert_dataclass_to_model(field_type))
    if isinstance(field_type, EnumMeta):
        enum_values: List[Union[str, int]] = [v.value for v in field_type]  # type: ignore
        openapi_type = (
            OpenAPIType.STRING if isinstance(enum_values[0], str) else OpenAPIType.INTEGER
        )
        return Schema(type=openapi_type, enum=enum_values)
    if field_type is UploadFile:
        return Schema(
            type=OpenAPIType.STRING,
            contentMediaType="application/octet-stream",
        )
    return Schema()


def create_examples_for_field(field: ModelField) -> List[Example]:
    try:
        value = clean_values_from_example(ExampleFactory.get_field_value(field))
        return [Example(description=f"Example {field.name} value", value=value)]
    except ParameterError:
        return []


def create_schema(field: ModelField, create_examples: bool, ignore_optional: bool = False):
    if is_optional(field) and not ignore_optional:
        non_optional_schema = create_schema(
            field=field, create_examples=False, ignore_optional=True
        )

        schema = Schema(
            oneOf=[
                Schema(type=OpenAPIType.NULL),
                *(
                    non_optional_schema.oneOf
                    if non_optional_schema.oneOf
                    else [non_optional_schema]
                ),
            ]
        )
    elif is_union(field):
        schema = Schema(
            oneOf=[
                create_schema(field=sub_field, create_examples=False)
                for sub_field in field.sub_fields or []
            ]
        )

    elif ModelFactory.is_constrained_field(field.type_):
        field.outer_type_ = field.type_
        schema = create_constrained_field_schema(
            field_type=field.outer_type_, sub_fields=field.sub_fields
        )
    elif field.sub_fields:
        openapi_type = get_openapi_type_for_complex_type(field)
        schema = Schema(type=openapi_type)
        if openapi_type == OpenAPIType.ARRAY:
            items = [
                create_schema(field=sub_field, create_examples=False)
                for sub_field in field.sub_fields
            ]
            if len(items) > 1:
                schema.items = Schema(oneOf=items)
            else:
                schema.items = items[0]
    else:
        schema = get_schema_for_field_type(field=field)
    if not ignore_optional:
        schema = update_schema_field_info(schema=schema, field_info=field.field_info)
    if not schema.examples and create_examples:
        schema.examples = create_examples_for_field(field=field)
    return schema



from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Generic,
    NoReturn,
    Optional,
    TypeVar,
    Union,
    cast,
)

from esmerald.enums import MediaType
from esmerald.exceptions import ImproperlyConfigured
from orjson import OPT_OMIT_MICROSECONDS  # noqa
from orjson import OPT_SERIALIZE_NUMPY, dumps
from pydantic import BaseModel
from starlette import status
from starlette.responses import FileResponse as FileResponse  # noqa
from starlette.responses import HTMLResponse as HTMLResponse  # noqa
from starlette.responses import JSONResponse as JSONResponse  # noqa
from starlette.responses import PlainTextResponse as PlainTextResponse  # noqa
from starlette.responses import RedirectResponse as RedirectResponse  # noqa
from starlette.responses import Response as StarletteResponse  # noqa
from starlette.responses import StreamingResponse as StreamingResponse  # noqa
from starlette.types import Receive, Scope, Send

if TYPE_CHECKING:
    from esmerald.backgound import BackgroundTask, BackgroundTasks
    from esmerald.protocols.template import TemplateEngineProtocol
    from esmerald.types import ResponseCookies

T = TypeVar("T")

try:
    import orjson
except ImportError:  # pragma: nocover
    orjson = None  # type: ignore

try:
    import ujson
except ImportError:  # pragma: nocover
    ujson = None  # type: ignore


class Response(StarletteResponse, Generic[T]):
    def __init__(
        self,
        content: T,
        *,
        status_code: Optional[int] = status.HTTP_200_OK,
        media_type: Optional[Union["MediaType", str]] = None,
        background: Optional[Union["BackgroundTask", "BackgroundTasks"]] = None,
        headers: Optional[Dict[str, Any]] = None,
        cookies: Optional["ResponseCookies"] = None,
    ) -> None:
        super().__init__(
            content=content,
            status_code=status_code,
            headers=headers or {},
            media_type=media_type,
            background=cast("BackgroundTask", background),
        )
        self.cookies = cookies or []

    @staticmethod
    def transform(value: Any) -> Dict[str, Any]:
        if isinstance(value, BaseModel):
            return value.dict()
        raise TypeError("unsupported type")  # pragma: no cover

    def render(self, content: Any) -> bytes:
        try:
            if (
                content is None
                or content is NoReturn
                and (
                    self.status_code < 100
                    or self.status_code
                    in {status.HTTP_204_NO_CONTENT, status.HTTP_304_NOT_MODIFIED}
                )
            ):
                return b""
            if self.media_type == MediaType.JSON:
                return dumps(
                    content,
                    default=self.transform,
                    option=OPT_SERIALIZE_NUMPY | OPT_OMIT_MICROSECONDS,
                )
            return super().render(content)
        except (AttributeError, ValueError, TypeError) as e:
            raise ImproperlyConfigured("Unable to serialize response content") from e


class BaseJSONResponse(JSONResponse):
    """
    Making sure it parses all the values from pydantic into dictionary.
    """

    @staticmethod
    def transform(value: Any) -> Dict[str, Any]:
        """
        Makes sure that every value is checked and if it's a pydantic model then parses into
        a dict().
        """
        if isinstance(value, BaseModel):
            return value.dict()
        raise TypeError("unsupported type")


class ORJSONResponse(BaseJSONResponse):
    def render(self, content: Any) -> bytes:
        assert orjson is not None, "orjson must be installed to use ORJSONResponse"
        return orjson.dumps(
            content,
            default=self.transform,
            option=OPT_SERIALIZE_NUMPY | OPT_OMIT_MICROSECONDS,
        )


class UJSONResponse(BaseJSONResponse):
    def render(self, content: Any) -> bytes:
        assert ujson is not None, "ujson must be installed to use UJSONResponse"
        return ujson.dumps(content, ensure_ascii=False, default=self.transform).encode("utf-8")


class TemplateResponse(Response):
    def __init__(
        self,
        template_name: str,
        template_engine: "TemplateEngineProtocol",
        status_code: int = 200,
        context: Optional[Dict[str, Any]] = None,
        background: Optional[Union["BackgroundTask", "BackgroundTasks"]] = None,
        headers: Optional[Dict[str, Any]] = None,
        cookies: Optional["ResponseCookies"] = None,
    ):
        self.template = template_engine.get_template(template_name)
        self.context = context or {}
        content = self.template.render(**context)
        super().__init__(
            content=content,
            status_code=status_code,
            headers=headers,
            media_type=MediaType.HTML,
            background=background,
            cookies=cookies,
        )

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        request = self.context.get("request", {})
        extensions = request.get("extensions", {})
        if "http.response.template" in extensions:
            await send(
                {
                    "type": "http.response.template",
                    "template": self.template,
                    "context": self.context,
                }
            )
        await super().__call__(scope, receive, send)



from typing import Any, TYPE_CHECKING, TypeVar, cast

from esmerald.exceptions import ImproperlyConfigured
from esmerald.typing import Void
from orjson import loads
from starlette.requests import ClientDisconnect as ClientDisconnect  # noqa
from starlette.requests import HTTPConnection as HTTPConnection  # noqa: F401
from starlette.requests import Request as StarletteRequest  # noqa: F401
from starlette.requests import empty_receive, empty_send  # noqa

if TYPE_CHECKING:
    from esmerald.applications import Esmerald
    from esmerald.types import HTTPMethod, Receive, Scope, Send

User = TypeVar("User")


class Request(StarletteRequest):
    def __init__(
        self,
        scope: "Scope",
        receive: "Receive" = empty_receive,
        send: "Send" = empty_send,
    ):
        super().__init__(scope, receive, send)
        self._json: Any = Void

    @property
    def app(self) -> "Esmerald":
        return cast("Esmerald", self.scope["app"])

    @property
    def user(self) -> User:
        if "user" not in self.scope:
            raise ImproperlyConfigured(
                "'user' is not defined in scope, install an AuthMiddleware to set it"
            )
        return cast("User", self.scope["user"])

    @property
    def method(self) -> "HTTPMethod":
        return cast("HTTPMethod", self.scope["method"])

    async def json(self):
        if self._json is Void:
            if "_body" in self.scope:
                body = self.scope["_body"]
            else:
                body = self.scope["_body"] = await self.body() or "null".encode("utf-8")
            self._json = loads(body)
        return self._json



from typing import Generic, TypeVar

from starlette.requests import HTTPConnection
from starlette.types import ASGIApp
from tortoise.exceptions import DoesNotExist

from esmerald.config import JWTConfig
from esmerald.exceptions import NotAuthorized
from esmerald.middleware.authentication import AuthResult, BaseAuthMiddleware
from esmerald.security.jwt.token import Token

T = TypeVar("T")


class JWTAuthMiddleware(BaseAuthMiddleware):
    """
    The simple JWT authentication Middleware.
    """

    def __init__(
        self,
        app: "ASGIApp",
        config: "JWTConfig",
        user_model: Generic[T],
    ):
        super().__init__(app)
        """
        The user is simply the class type to be queried from the Tortoise ORM.

        Example how to use:

            1. User table

                from esmerald.contrib.auth.tortoise.base_user import User as BaseUser

                class User(BaseUser):
                    ...

            2. Middleware

                from esmerald.contrib.auth.tortoise.middleware import JWTAuthMiddleware
                from esmerald.config import JWTConfig

                jwt_config = JWTConfig(...)

                class CustomJWTMidleware(JWTAuthMiddleware):
                    def __init__(self, app: "ASGIApp"):
                        super().__init__(app, config=jwt_config, user=User)

            3. The application
                from esmerald import Esmerald
                from myapp.middleware import CustomJWTMidleware

                app = Esmerald(routes=[...], middleware=[CustomJWTMidleware])

        """
        self.app = app
        self.config = config
        self.user_model = user_model

    async def retrieve_user(self, user_id: int) -> Generic[T]:
        """
        Retrieves a user from the database using the given token id.
        """
        try:
            await self.user_model.get(pk=user_id)
        except DoesNotExist:
            raise NotAuthorized()

    async def authenticate(self, request: HTTPConnection) -> AuthResult:
        token = request.headers.get(self.config.api_key_header)

        if not token:
            raise NotAuthorized("JWT token not found.")

        token = Token.decode(
            token=token, key=self.config.signing_key, algorithm=self.config.algorithm
        )

        user = await self.retrieve_user(token.sub)
        return AuthResult(user=user)



from typing import Any, Dict, List, Optional, TypeVar, Union

from pydantic import DirectoryPath, validate_arguments
from typing_extensions import Protocol, runtime_checkable


@runtime_checkable
class TemplateProtocol(Protocol):  # pragma: no cover
    def render(self, **context: Optional[Dict[str, Any]]) -> str:
        ...


TP = TypeVar("TP", bound=TemplateProtocol, covariant=True)


@runtime_checkable
class TemplateEngineProtocol(Protocol[TP]):  # pragma: no cover
    @validate_arguments
    def __init__(self, directory: Union[DirectoryPath, List[DirectoryPath]]) -> None:
        ...

    def get_template(self, template_name: str) -> TP:
        ...



from datetime import datetime
from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Callable,
    Coroutine,
    Dict,
    List,
    Type,
    TypeVar,
    Union,
)

from esmerald.backgound import BackgroundTask, BackgroundTasks
from esmerald.exceptions import HTTPException, StarletteHTTPException
from esmerald.routing.gateways import WebSocketGateway
from esmerald.routing.router import Include
from esmerald.schedulers import AsyncIOScheduler
from pydantic.typing import AnyCallable
from starlette.responses import Response as StarletteResponse
from typing_extensions import Literal

if TYPE_CHECKING:
    from esmerald.datastructures import Cookie, ResponseHeader, State  # noqa: TC004
    from esmerald.injector import Inject  # noqa
    from esmerald.protocols.middleware import MiddlewareProtocol
    from esmerald.requests import Request  # noqa
    from esmerald.responses import Response  # noqa
    from esmerald.routing.router import Gateway, HTTPHandler, Router  # noqa
    from esmerald.routing.views import APIView  # noqa
    from esmerald.websockets import WebSocket  # noqa
    from starlette.middleware import Middleware as StarletteMiddleware  # noqa: TC004
    from starlette.middleware.base import BaseHTTPMiddleware  # noqa: TC004
    from starlette.types import ASGIApp as ASGIApp  # noqa
    from starlette.types import Message as Message  # noqa
    from starlette.types import Receive as Receive  # noqa
    from starlette.types import Scope as Scope  # noqa
    from starlette.types import Send as Send  # noqa
else:
    ASGIApp = Any
    HTTPHandler = Any
    Message = Any
    Receive = Any
    Scope = Any
    Send = Any
    StarletteMiddleware = Any
    BaseHTTPMiddleware = Any
    Inject = Any
    Request = Any
    WebSocket = Any
    State = Any
    Response = Any
    ResponseHeader = Any
    Cookie = Any
    Router = Any
    MiddlewareProtocol = Any
    APIView = Any
    Gateway = Any

AsyncAnyCallable = Callable[..., Awaitable[Any]]
HTTPMethod = Literal["GET", "POST", "DELETE", "PATCH", "PUT", "HEAD"]

Middleware = Union[
    StarletteMiddleware,
    Type[BaseHTTPMiddleware],
    Type[MiddlewareProtocol],
    Callable[..., ASGIApp],
]

ResponseType = Type[Response]

Dependencies = Dict[str, Inject]

CoroutineHandler = Coroutine[Any, Any, Response]
ExceptionHandler = Union[
    Callable[
        [Request, Union[Exception, HTTPException, StarletteHTTPException]],
        StarletteResponse,
    ],
    Callable[[Response, Any], CoroutineHandler],
]
ExceptionHandlers = Dict[Union[int, Type[Exception]], ExceptionHandler]

ReservedKwargs = Literal[
    "request",
    "socket",
    "headers",
    "query",
    "cookies",
    "state",
    "data",
]

ResponseHeaders = Dict[str, ResponseHeader]
ResponseCookies = List[Cookie]
AsyncAnyCallable = Callable[..., Awaitable[Any]]

LifeSpanHandler = Union[
    Callable[[], Any],
    Callable[[State], Any],
    Callable[[], Awaitable[Any]],
    Callable[[State], Awaitable[Any]],
]

SchedulerType = AsyncIOScheduler
DatetimeType = TypeVar("DatetimeType", bound=datetime)

ParentType = Union[APIView, Router]
APIGateHandler = Union[
    Type[APIView],
    HTTPHandler,
    Router,
    AnyCallable,
    Gateway,
    WebSocketGateway,
]

RouteParent = Union["Router", "Include", "ASGIApp", "Gateway", "WebSocketGateway"]

BackgroundTaskType = Union[BackgroundTask, BackgroundTasks]
SecurityRequirement = Dict[str, List[str]]

ConnectionType = Union["Request", "WebSocket"]
DictStr = Dict[str, str]



from functools import partial
from typing import Awaitable, Callable, Generic, List, TypeVar, Union

from anyio.to_thread import run_sync
from esmerald.utils.helpers import is_async_callable
from typing_extensions import ParamSpec

P = ParamSpec("P")
T = TypeVar("T")


class AsyncCallable(Generic[P, T]):
    __slots__ = ("args", "kwargs", "fn")

    def __init__(self, fn: Callable[P, T]):
        self.fn: Callable[P, Awaitable[T]]
        if is_async_callable(fn):
            self.fn = fn
        else:
            self.fn = partial(run_sync, fn)  # type:ignore[assignment]

    async def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T:
        return await self.fn(*args, **kwargs)


def as_async_callable_list(value: Union[Callable, List[Callable]]) -> List[AsyncCallable]:
    if not isinstance(value, list):
        return [AsyncCallable(value)]
    return [AsyncCallable(v) for v in value]



import os
from typing import Any, Optional, Type

from esmerald.utils.functional import LazyObject, empty
from esmerald.utils.module_loading import import_string

ENVIRONMENT_VARIABLE = "ESMERALD_SETTINGS_MODULE"

APISettings = Type["EsmeraldLazySettings"]


class EsmeraldLazySettings(LazyObject):
    """
    A lazy proxy for either global Esmerald settings or a custom settings object.
    The user can manually configure settings prior to using them. Otherwise,
    Esmerald uses the settings module pointed to by ESMERALD_SETTINGS_MODULE.
    """

    def _setup(self, name: Optional[str] = None) -> None:
        """
        Load the settings module pointed to by the environment variable. This
        is used the first time settings are needed, if the user hasn't
        configured settings manually.
        """
        settings_module: str = os.environ.get(
            ENVIRONMENT_VARIABLE, "esmerald.conf.global_settings.EsmeraldAPISettings"
        )

        settings: Any = import_string(settings_module)

        for setting, _ in settings().dict().items():
            assert setting.islower(), "%s should be in lowercase." % setting

        self._wrapped = settings()

    def __repr__(self: "EsmeraldLazySettings"):
        # Hardcode the class name as otherwise it yields 'Settings'.
        if self._wrapped is empty:
            return "<EsmeraldLazySettings [Unevaluated]>"
        return '<EsmeraldLazySettings "%(settings_module)s">' % {
            "settings_module": self._wrapped.__class__.__name__
        }

    @property
    def configured(self):
        """Return True if the settings have already been configured."""
        return self._wrapped is not empty


settings: APISettings = EsmeraldLazySettings()



from typing import (
    TYPE_CHECKING,
    Any,
    AsyncContextManager,
    Callable,
    Dict,
    List,
    Optional,
    Union,
    cast,
)

import httpx  # noqa
from esmerald.applications import Esmerald
from esmerald.conf import settings  # noqa
from esmerald.utils.crypto import get_random_secret_key
from starlette.testclient import TestClient  # noqa

if TYPE_CHECKING:
    from esmerald.config import (
        CORSConfig,
        CSRFConfig,
        OpenAPIConfig,
        SessionConfig,
        StaticFilesConfig,
        TemplateConfig,
    )
    from esmerald.permissions.types import Permission
    from esmerald.types import (
        APIGateHandler,
        Dependencies,
        ExceptionHandlers,
        LifeSpanHandler,
        Middleware,
        SchedulerType,
        DictStr,
    )
    from typing_extensions import Literal


class EsmeraldTestClient(TestClient):
    app: Esmerald

    def __init__(
        self,
        app: Esmerald,
        base_url: str = "http://testserver",
        raise_server_exceptions: bool = True,
        root_path: str = "",
        backend: "Literal['asyncio', 'trio']" = "asyncio",
        backend_options: Optional[Dict[str, Any]] = None,
        cookies: Optional[httpx._client.CookieTypes] = None,
        headers: Optional["DictStr"] = None,
    ):
        super().__init__(
            app=app,
            base_url=base_url,
            raise_server_exceptions=raise_server_exceptions,
            root_path=root_path,
            backend=backend,
            backend_options=backend_options,
            cookies=cookies,
            headers=headers,
        )

    def __enter__(self, *args: Any, **kwargs: Dict[str, Any]) -> "EsmeraldTestClient":
        return super().__enter__(*args, **kwargs)


def create_client(
    routes: Union["APIGateHandler", List["APIGateHandler"]],
    *,
    debug: bool = settings.debug,
    app_name: Optional[str] = settings.app_name,
    secret_key: Optional[str] = get_random_secret_key(),
    allowed_hosts: Optional[List[str]] = settings.allowed_hosts,
    allow_origins: Optional[List[str]] = settings.allow_origins,
    base_url: str = "http://testserver",
    backend: "Literal['asyncio', 'trio']" = "asyncio",
    backend_options: Optional[Dict[str, Any]] = None,
    permissions: Optional[List["Permission"]] = settings.permissions,
    dependencies: Optional["Dependencies"] = settings.dependencies,
    middleware: Optional[List["Middleware"]] = settings.middleware,
    csrf_config: Optional["CSRFConfig"] = settings.csrf_config,
    exception_handlers: Optional["ExceptionHandlers"] = settings.exception_handlers,
    openapi_config: Optional["OpenAPIConfig"] = settings.openapi_config,
    on_shutdown: Optional[List["LifeSpanHandler"]] = settings.on_shutdown,
    on_startup: Optional[List["LifeSpanHandler"]] = settings.on_startup,
    cors_config: Optional["CORSConfig"] = settings.cors_config,
    session_config: Optional["SessionConfig"] = settings.session_config,
    scheduler_class: Optional["SchedulerType"] = settings.scheduler_class,
    scheduler_tasks: Optional[Dict[str, str]] = settings.scheduler_tasks,
    scheduler_configurations: Optional[
        Dict[str, Union[str, Dict[str, str]]]
    ] = settings.scheduler_configurations,
    enable_scheduler: bool = settings.enable_scheduler,
    raise_server_exceptions: bool = True,
    root_path: str = "",
    static_files_config: Optional[Union["StaticFilesConfig", List["StaticFilesConfig"]]] = None,
    template_config: Optional["TemplateConfig"] = None,
    lifespan: Optional[Callable[["Esmerald"], "AsyncContextManager"]] = settings.lifespan,
    cookies: Optional[httpx._client.CookieTypes] = None,
    redirect_slashes: bool = settings.redirect_slashes,
) -> EsmeraldTestClient:
    return EsmeraldTestClient(
        app=Esmerald(
            debug=debug,
            routes=cast("Any", routes if isinstance(routes, list) else [routes]),
            app_name=app_name,
            secret_key=secret_key,
            allowed_hosts=allowed_hosts,
            allow_origins=allow_origins,
            permissions=permissions,
            dependencies=dependencies,
            middleware=middleware,
            csrf_config=csrf_config,
            exception_handlers=exception_handlers,
            openapi_config=openapi_config,
            on_shutdown=on_shutdown,
            on_startup=on_startup,
            cors_config=cors_config,
            scheduler_class=scheduler_class,
            scheduler_tasks=scheduler_tasks,
            scheduler_configurations=scheduler_configurations,
            enable_scheduler=enable_scheduler,
            static_files_config=static_files_config,
            template_config=template_config,
            session_config=session_config,
            lifespan=lifespan,
            redirect_slashes=redirect_slashes,
        ),
        base_url=base_url,
        backend=backend,
        backend_options=backend_options,
        root_path=root_path,
        raise_server_exceptions=raise_server_exceptions,
        cookies=cookies,
    )



from datetime import datetime, timezone
from typing import TYPE_CHECKING, Dict, List, Optional, Union

from esmerald.exceptions import ImproperlyConfigured
from esmerald.security.utils import convert_time
from jose import JWSError, JWTError, jwt
from pydantic import BaseModel, Field, constr, validator

if TYPE_CHECKING:
    from pydantic.typing import DictAny


class Token(BaseModel):
    """
    Classic representation of a token via pydantic model.
    """

    exp: datetime
    iat: datetime = Field(default_factory=lambda: convert_time(datetime.now(timezone.utc)))
    sub: constr(min_length=1)
    iss: Optional[str] = None
    aud: Optional[str] = None
    jti: Optional[str] = None

    @validator("exp", always=True)
    def validate_expiration(cls, date: datetime) -> datetime:
        """
        When a token is issued, needs to be date in the future.
        """
        date = convert_time(date)
        if date.timestamp() >= convert_time(datetime.now(timezone.utc)).timestamp():
            return date
        raise ValueError("The exp must be a date in the future.")

    @validator("iat", always=True)
    def validate_iat(cls, date: datetime) -> datetime:  # pylint: disable=no-self-argument
        """Ensures that the `Issued At` it's nt bigger than the current time."""
        date = convert_time(date)
        if date.timestamp() <= convert_time(datetime.now(timezone.utc)).timestamp():
            return date
        raise ValueError("iat must be a current or past time")

    @staticmethod
    def encode(key: str, algorithm: str, **claims: "DictAny") -> str:
        """
        Encodes the token into a proper str formatted and allows passing kwargs.
        """
        try:
            return jwt.encode(
                claims=claims,
                key=key,
                algorithm=algorithm,
            )
        except (JWSError, JWTError) as e:
            raise ImproperlyConfigured("Error encoding the token.") from e

    @staticmethod
    def decode(token: str, key: Union[str, Dict[str, str]], algorithms: List[str]) -> "Token":
        """
        Decodes the given token.
        """
        try:
            data = jwt.decode(
                token=token,
                key=key,
                algorithms=[algorithms],
                options={"verify_aud": False},
            )
        except (JWSError, JWTError) as e:
            raise ImproperlyConfigured("Invalid token.") from e
        return Token(**data)



from typing import (
    TYPE_CHECKING,
    Any,
    AsyncContextManager,
    Callable,
    Dict,
    List,
    Optional,
    Union,
)

from esmerald import __version__
from esmerald.conf.enums import EnvironmentType
from esmerald.config import (
    CORSConfig,
    CSRFConfig,
    OpenAPIConfig,
    SessionConfig,
    StaticFilesConfig,
    TemplateConfig,
)
from esmerald.config.asyncexit import AsyncExitConfig
from esmerald.permissions.types import Permission
from esmerald.schedulers import AsyncIOScheduler
from esmerald.types import (
    APIGateHandler,
    Dependencies,
    ExceptionHandlers,
    LifeSpanHandler,
    Middleware,
    ResponseCookies,
    ResponseHeaders,
    ResponseType,
    SchedulerType,
)
from openapi_schemas_pydantic.v3_1_0 import License, SecurityRequirement, Server
from pydantic import BaseSettings

if TYPE_CHECKING:
    from esmerald.applications import Esmerald


class EsmeraldAPISettings(BaseSettings):
    debug: bool = False
    environment: Optional[str] = EnvironmentType.PRODUCTION
    app_name: str = "Esmerald"
    title: str = "My awesome Esmerald application"
    description: str = "Highly scalable, performant, easy to learn and for every application."
    contact: Optional[Dict[str, Union[str, Any]]] = {
        "name": "admin",
        "email": "admin@myapp.com",
    }
    summary: str = "Esmerald application"
    terms_of_service: Optional[str] = None
    license: Optional[License] = None
    security: Optional[List[SecurityRequirement]] = None
    servers: List[Server] = [Server(url="/")]
    secret_key: str = "my secret"
    version: str = __version__
    allowed_hosts: Optional[List[str]] = ["*"]
    allow_origins: Optional[List[str]] = None
    response_class: Optional[ResponseType] = None
    response_cookies: Optional[ResponseCookies] = None
    response_headers: Optional[ResponseHeaders] = None
    scheduler_class: SchedulerType = AsyncIOScheduler
    include_in_schema: bool = True
    tags: Optional[List[str]] = None
    timezone: str = "UTC"
    use_tz: bool = False
    root_path: Optional[str] = ""
    enable_sync_handlers: bool = True
    enable_scheduler: bool = False
    enable_openapi: bool = True
    redirect_slashes: bool = True

    @property
    def reload(self) -> bool:
        """
        Returns reloading for dev and test environments.
        """
        if self.environment in [EnvironmentType.DEVELOPMENT, EnvironmentType.TESTING]:
            return True
        return False

    @property
    def password_hashers(self) -> List[str]:
        return [
            "esmerald.contrib.auth.hashers.PBKDF2PasswordHasher",
            "esmerald.contrib.auth.hashers.PBKDF2SHA1PasswordHasher",
        ]

    @property
    def routes(self) -> List[APIGateHandler]:
        """
        Property that can be used as an entrypoint for the base app routes and to start the application.

        Example:
            from esmerald import Include

            class MySettings(EsmeraldAPISettings):
                @property
                def routes(self):
                    return Include(path='/api/v1/', namespace='myapp.routes')
        """
        return []

    @property
    def csrf_config(self) -> CSRFConfig:
        """
        Initial Default configuration for the CSRF.
        This can be overwritten in another setting or simply override `secret` or then override
        the `def csrf_config()` property to change the behavior of the whole csrf_config.

        Default:
            None

        Example:

            class MySettings(EsmeraldAPISettings):
                secret: str = "n(0t=%_amauq1m&6sde4z#3mkdmfcad1942ny&#sjp1oygk-5_"

                @property
                def csrf_config(self) -> CSRFConfig:
                    if not self.secret_key:
                        raise ImproperlyConfigured("`secret` setting not configured.")
                    return CSRFConfig(secret=self.secret_key)
        """
        return None

    @property
    def async_exit_config(self) -> AsyncExitConfig:
        """
        Initial Default configuration for the CSRF.
        This can be overwritten in another setting or simply override `secret` or then override
        the `def async_exit_config()` property to change the behavior of the whole async_exit_config.

        This replaces the classic:
        `async_exit_config: Optional[AsyncExitConfig] = None`.

        Default:
            AsyncExitConfig

        Example:

            class MySettings(EsmeraldAPISettings):
                @property
                def async_exit_config(self) -> AsyncExitConfig:
                    ...
        """
        return AsyncExitConfig()

    @property
    def template_config(self) -> TemplateConfig:
        """
        Initial Default configuration for the TemplateConfig.
        This can be overwritten in another setting or simply override `template_config` or then override
        the `def template_config()` property to change the behavior of the whole template_config.

        Esmerald can also support other engines like mako, Diazo, Cheetah. Currently natively
        only supports jinja2 and mako.

        Default:
            JinjaTemplateEngine

        Example:

            class MySettings(EsmeraldAPISettings):
                @property
                def template_config(self) -> TemplateConfig:
                    TemplateConfig(directory='templates', engine=MakoTemplateEngine)
        """
        return None

    @property
    def static_files_config(self) -> StaticFilesConfig:
        """
        Simple configuration indicating where the statics will be placed in
        the application.

        Default:
            None

        Example:

            class MySettings(EsmeraldAPISettings):
                @property
                def static_files_config(self) -> StaticFilesConfig:
                    StaticFilesConfig(path='/', directories=...)
        """
        return None

    @property
    def cors_config(self) -> CORSConfig:
        """
        Initial Default configuration for the CORS.
        This can be overwritten in another setting or simply override `allow_origins` or then override
        the `def cors_config()` property to change the behavior of the whole cors_config.

        Default:
            CORSConfig

        Example:

            class MySettings(EsmeraldAPISettings):
                allow_origins: List[str] = ['www.example.com', 'www.foobar.com']

                @property
                def cors_config(self) -> CORSConfig:
                    ...
        """
        if not self.allow_origins:
            return None
        return CORSConfig(allow_origins=self.allow_origins)

    @property
    def session_config(self) -> SessionConfig:
        """
        Initial Default configuration for the SessionConfig.
        This can be overwritten in another setting or simply override `session_config` or then override
        the `def session_config()` property to change the behavior of the whole session_config.

        Default:
            None

        Example:

            class MySettings(EsmeraldAPISettings):
                @property
                def session_config(self) -> SessionConfig:
                    SessionConfig(engine=MakoTemplateEngine)
        """
        return None

    @property
    def openapi_config(self) -> OpenAPIConfig:
        """
        Initial Default configuration for the OpenAPI.
        This can be overwritten in another setting or simply override `openapi_config` or then override the `def openapi_config()` property to change the behavior.

        Default:
            OpenAPIConfig

        Example:

            class MySettings(EsmeraldAPISettings):

                @property
                def openapi_config(self) -> OpenAPIConfig:
                    ...
        """
        from esmerald.openapi.apiview import OpenAPIView

        return OpenAPIConfig(
            openapi_apiview=OpenAPIView,
            title=self.title,
            version=self.version,
            contact=self.contact,
            description=self.description,
            terms_of_service=self.terms_of_service,
            license=self.license,
            servers=self.servers,
            summary=self.summary,
            security=self.security,
            tags=self.tags,
        )

    @property
    def middleware(self) -> List[Middleware]:
        """
        Initial Default configuration for the middleware.
        This can be overwritten in another setting or simply override `def middleware()`.

        Example:

            class MySettings(EsmeraldAPISettings):

                @property
                def middleware(self) -> List[Middleware]:
                    return [EsmeraldMiddleware]
        """
        return None

    @property
    def scheduler_tasks(self) -> Dict[str, str]:
        """Returns a dict of tasks for run with `scheduler_class`.

        Where the tasks are placed is not linked to the name of
        the file itself. They can be anywhere. What is imoprtant
        is that in the dictionary the name of the task and the
        location of the file where the task is.

        Returns:
            Dict[str, str]: A list of tasks.

        Example:

            class MySettings(EsmeraldAPISettings):

                @property
                def scheduler_tasks(self) -> Dict[str, str]:
                    tasks = {
                        "send_newslettters": "accounts.tasks",
                        "check_balances": "finances.balance_tasks",
                    }

        """

        return {}

    @property
    def scheduler_configurations(self) -> Dict[str, Union[str, Dict[str, str]]]:
        """Returns a dict of configurations for run with `scheduler_class`.

        Example:

            class MySettings(EsmeraldAPISettings):

                @property
                def scheduler_configurations(self) -> Dict[str, str]:
                    configurations = {
                        'apscheduler.jobstores.mongo': {
                            'type': 'mongodb'
                        },
                        'apscheduler.jobstores.default': {
                            'type': 'sqlalchemy',
                            'url': 'sqlite:///jobs.sqlite'
                        },
                        'apscheduler.executors.default': {
                            'class': 'apscheduler.executors.pool:ThreadPoolExecutor',
                            'max_workers': '20'
                        },
                        'apscheduler.executors.processpool': {
                            'type': 'processpool',
                            'max_workers': '5'
                        },
                        'apscheduler.job_defaults.coalesce': 'false',
                        'apscheduler.job_defaults.max_instances': '3',
                        'apscheduler.timezone': 'UTC',
                    }
        """

        return {}

    @property
    def permissions(self) -> List[Permission]:
        """
        Returns the default permissions of Esmerald.
        """
        return None

    @property
    def dependencies(self) -> Dependencies:
        """
        Returns the dependencies of Esmerald main app.
        """
        return None

    @property
    def exception_handlers(self) -> ExceptionHandlers:
        """
        Default exception handlers to be loaded when the application starts
        """
        return None

    @property
    def on_startup(self) -> List[LifeSpanHandler]:
        """
        List of events/actions to be done on_startup.
        """
        return None

    @property
    def on_shutdown(self) -> List[LifeSpanHandler]:
        """
        List of events/actions to be done on_shutdown.
        """
        return None

    @property
    def lifespan(self) -> Callable[["Esmerald"], AsyncContextManager]:
        """
        Custom lifespan that can be passed instead of the default Starlette.

        The lifespan context function is a newer style that replaces
        on_startup / on_shutdown handlers. Use one or the other, not both.
        """
        return None



from argparse import HelpFormatter
from io import TextIOBase
from typing import Any, Dict, Optional


class EsmeraldHelpFormatter(HelpFormatter):
    show_last = {
        "--version",
        "--verbosity",
        "--traceback",
    }

    def _reordered_actions(self, actions: Any):
        return sorted(actions, key=lambda a: set(a.option_strings) & self.show_last != set())

    def add_usage(self, usage: str, actions: Any, *args: Any, **kwargs: Dict[str, Any]):
        super().add_usage(usage, self._reordered_actions(actions), *args, **kwargs)

    def add_arguments(self, actions: Any):
        super().add_arguments(self._reordered_actions(actions))


class OutputWrapper(TextIOBase):
    def __init__(self, out: Any, ending: str = "\n"):
        self._out = out
        self.ending = ending

    def __getattr__(self, name: str):
        return getattr(self._out, name)

    def flush(self) -> Any:
        if hasattr(self._out, "flush"):
            self._out.flush()

    def write(
        self,
        msg: str = "",
        ending: Optional[Any] = None,
    ) -> Any:
        ending = self.ending if ending is None else ending
        if ending and not msg.endswith(ending):
            msg += ending
        self._out.write(msg)



from typing import Generic, Optional, TypeVar

from starlette.requests import HTTPConnection as HTTPConnection  # noqa: F401
from starlette.websockets import WebSocket as StarletteWebSocket  # noqa
from starlette.websockets import WebSocketClose as WebSocketClose  # noqa
from starlette.websockets import (
    WebSocketDisconnect as StarletteWebSocketDisconnect,
)  # noqa
from starlette.websockets import WebSocketState as WebSocketState  # noqa

User = TypeVar("User")
Auth = TypeVar("Auth")


class WebSocket(StarletteWebSocket, Generic[User, Auth]):
    ...


class WebSocketDisconnect(StarletteWebSocketDisconnect):
    """Esmerald WebSocketDisconnect"""

    def __init__(self, code: int = 1000, reason: Optional[str] = None) -> None:
        super().__init__(code, reason)



from esmerald.core.management.templates import TemplateDirective


class Directive(TemplateDirective):
    help = (
        "Creates an Esmerald app directory structure for the given app name in "
        "the current directory or optionally in the given directory."
    )
    missing_args_message = "You must provide an application name."

    def handle(self, **options):
        app_name = options.pop("name")
        super().handle("app", app_name, **options)
